# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import List, Union

from servicex.models import TransformRequest


class DataSetIdentifier:
    r"""
    Base class for specifying the dataset to transform. This can either be a list of
    xRootD URIs or a rucio DID
    """
    def __init__(self, scheme: str, dataset: str, num_files: int = None):
        self.scheme = scheme
        self.dataset = dataset
        self.num_files = num_files

    @property
    def did(self):
        num_files_arg = f"?files={self.num_files}" if self.num_files else ""
        return f"{self.scheme}://{self.dataset}{num_files_arg}"

    def populate_transform_request(self, transform_request: TransformRequest) -> None:
        transform_request.did = self.did
        transform_request.file_list = None


class RucioDatasetIdentifier(DataSetIdentifier):
    def __init__(self, dataset: str, num_files: int = None):
        r"""
        Rucio Dataset - this will be looked up using the Rucio data management
        service.

        :param dataset: The rucio DID - this can be a dataset or a container of datasets.
        :param num_files: Maximum number of files to return. This is useful during development
            to perform quick runs. ServiceX is careful to make sure it always
            returns the same subset of files.

        """
        super().__init__("rucio", dataset, num_files=num_files)


class FileListDataset:
    def __init__(self, files: Union[List[str], str]):
        r"""
        Dataset specified as a list of XRootD URIs.

        :param files: Either a list of URIs or a single URI string
        """
        if type(files) is str:
            self.files = [files]
        else:
            self.files = files

    def populate_transform_request(self, transform_request: TransformRequest) -> None:
        transform_request.file_list = self.files
        transform_request.did = None



import pytest
import pandas as pd
import tempfile
import os

from unittest.mock import AsyncMock, Mock
from servicex.dataset_identifier import FileListDataset
from servicex.configuration import Configuration
from servicex.minio_adapter import MinioAdapter
from servicex.python_dataset import PythonQuery
from servicex.query_cache import QueryCache
from servicex.expandable_progress import ExpandableProgress

from pathlib import Path
from servicex.models import (
    ResultFormat,
    Status,
)
from rich.progress import Progress


@pytest.mark.asyncio
async def test_as_signed_urls_happy(transformed_result):
    # Test when display_progress is True and provided_progress is None
    did = FileListDataset("/foo/bar/baz.root")
    dataset = PythonQuery(dataset_identifier=did, codegen="uproot",
                          sx_adapter=None, query_cache=None)
    dataset.submit_and_download = AsyncMock()
    dataset.submit_and_download.return_value = transformed_result

    result = dataset.as_signed_urls(display_progress=True, provided_progress=None)
    assert result == transformed_result


@pytest.mark.asyncio
async def test_as_signed_urls_happy_dataset_group(transformed_result):
    # Test when display_progress is True and provided_progress is None
    did = FileListDataset("/foo/bar/baz.root")
    dataset = PythonQuery(dataset_identifier=did, codegen="uproot",
                          sx_adapter=None, query_cache=None)
    dataset.submit_and_download = AsyncMock()
    dataset.submit_and_download.return_value = transformed_result

    result = dataset.as_signed_urls(display_progress=True, provided_progress=None,
                                    dataset_group=True)
    assert result == transformed_result


@pytest.mark.asyncio
async def test_as_files_happy(transformed_result):
    did = FileListDataset("/foo/bar/baz.root")
    dataset = PythonQuery(dataset_identifier=did, codegen="uproot",
                          sx_adapter=None, query_cache=None)
    dataset.submit_and_download = AsyncMock()
    dataset.submit_and_download.return_value = transformed_result

    result = dataset.as_files(display_progress=True, provided_progress=None)
    assert result == transformed_result


@pytest.mark.asyncio
async def test_as_pandas_happy(transformed_result):
    did = FileListDataset("/foo/bar/baz.root")
    servicex = AsyncMock()
    with tempfile.TemporaryDirectory() as temp_dir:
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        dataset = PythonQuery(dataset_identifier=did, codegen="uproot", sx_adapter=servicex,
                              query_cache=cache)
        dataset.submit_and_download = AsyncMock()
        dataset.submit_and_download.return_value = transformed_result
        result = dataset.as_pandas(display_progress=False)
        assert isinstance(result, pd.DataFrame)
        assert not result.empty  # Check if the DataFrame is not empty
        assert dataset.result_format == ResultFormat.parquet
        cache.close()


@pytest.mark.asyncio
async def test_as_files_cached(transformed_result, python_dataset):
    python_dataset.cache = AsyncMock()
    python_dataset.cache.get_transform_by_hash = Mock()
    python_dataset.cache.get_transform_by_hash.return_value = transformed_result
    result = python_dataset.as_files(display_progress=True, provided_progress=None)
    assert result.title == transformed_result.title
    assert result.request_id == transformed_result.request_id
    assert result.files == transformed_result.files
    assert result.file_list == transformed_result.file_list
    assert result.codegen == transformed_result.codegen
    assert result.data_dir == transformed_result.data_dir


@pytest.mark.asyncio
async def test_download_files(python_dataset, transformed_result):
    signed_urls_only = False
    download_progress = "download_task_id"
    minio_mock = AsyncMock()
    config = Configuration(cache_path="temp_dir", api_endpoints=[])
    python_dataset.configuration = config
    minio_mock.download_file.return_value = Path("/path/to/downloaded_file")
    minio_mock.get_signed_url.return_value = Path("http://example.com/signed_url")
    minio_mock.list_bucket.return_value = [Mock(filename="file1.txt"),
                                           Mock(filename="file2.txt")]

    progress_mock = Mock()
    cached_record_mock = Mock(spec=transformed_result)
    python_dataset.minio_polling_interval = 0
    python_dataset.minio = minio_mock
    python_dataset.current_status = Mock(status="complete")
    python_dataset.configuration.shortened_downloaded_filename = False

    result_uris = await python_dataset.download_files(
        signed_urls_only, progress_mock, download_progress, cached_record_mock
    )
    minio_mock.download_file.assert_called()
    minio_mock.get_signed_url.assert_not_called()
    assert result_uris == ["/path/to/downloaded_file", "/path/to/downloaded_file"]


@pytest.mark.asyncio
async def test_download_files_with_signed_urls(python_dataset, transformed_result):
    signed_urls_only = True
    download_progress = "download_task_id"
    minio_mock = AsyncMock()
    config = Configuration(cache_path="temp_dir", api_endpoints=[])
    python_dataset.configuration = config
    minio_mock.download_file.return_value = "/path/to/downloaded_file"
    minio_mock.get_signed_url.return_value = "http://example.com/signed_url"
    minio_mock.list_bucket.return_value = [Mock(filename="file1.txt"),
                                           Mock(filename="file2.txt")]
    progress_mock = Mock()
    cached_record_mock = Mock(spec=transformed_result)

    python_dataset.minio_polling_interval = 0
    python_dataset.minio = minio_mock
    python_dataset.current_status = Mock(status="complete")
    python_dataset.configuration.shortened_downloaded_filename = False

    result_uris = await python_dataset.download_files(
        signed_urls_only, progress_mock, download_progress, cached_record_mock
    )
    minio_mock.download_file.assert_not_called()
    minio_mock.get_signed_url.assert_called()
    assert result_uris == ["http://example.com/signed_url", "http://example.com/signed_url"]


@pytest.mark.asyncio
async def test_transform_status_listener_happy(python_dataset):
    progress = Mock(spec=Progress)
    progress_task = Mock()
    download_task = Mock()
    status = Mock(files=10, files_completed=5, files_failed=1, status=Status.complete)
    python_dataset.current_status = status
    python_dataset.retrieve_current_transform_status = AsyncMock(return_value=status)
    await python_dataset.transform_status_listener(progress, progress_task, "mock_title",
                                                   download_task, "mock_title")

    python_dataset.retrieve_current_transform_status.assert_awaited_once()
    # progress.update.assert_called_with(progress_task, total=10)
    # progress.start_task.assert_called_with(progress_task)
    # progress.update.assert_called_with(progress_task, completed=5)
    assert python_dataset.files_completed == 5
    assert python_dataset.files_failed == 1


@pytest.mark.asyncio
async def test_retrieve_current_transform_status_status_none(python_dataset, completed_status):
    with tempfile.TemporaryDirectory() as temp_dir:
        python_dataset.current_status = None
        python_dataset.servicex = AsyncMock()
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        python_dataset.cache = cache
        python_dataset.configuration = config
        python_dataset.servicex.get_transform_status.return_value = completed_status

        await python_dataset.retrieve_current_transform_status()
        assert python_dataset.current_status == completed_status
        result = Path(os.path.join(Path(python_dataset.configuration.cache_path),
                      completed_status.request_id))
        python_dataset.download_path = result
        assert python_dataset.minio is not None
        assert isinstance(python_dataset.minio, MinioAdapter)
        cache.close()


@pytest.mark.asyncio
async def test_retrieve_current_transform_status_status_not(python_dataset, completed_status):
    with tempfile.TemporaryDirectory() as temp_dir:
        python_dataset.servicex = AsyncMock()
        python_dataset.servicex.get_transform_status.return_value = completed_status
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        python_dataset.cache = cache
        python_dataset.configuration = config

        await python_dataset.retrieve_current_transform_status()
        assert python_dataset.current_status == completed_status
        result = Path(os.path.join(Path(python_dataset.configuration.cache_path),
                      completed_status.request_id))
        python_dataset.download_path = result
        assert python_dataset.minio is not None
        assert isinstance(python_dataset.minio, MinioAdapter)
        cache.close()


@pytest.mark.asyncio
async def test_submit_and_download_cache_miss(python_dataset, completed_status):
    with tempfile.TemporaryDirectory() as temp_dir:
        python_dataset.current_status = None
        python_dataset.servicex = AsyncMock()
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        python_dataset.cache = cache
        python_dataset.configuration = config

        python_dataset.servicex = AsyncMock()
        python_dataset.cache.get_transform_by_hash = Mock()
        python_dataset.cache.get_transform_by_hash.return_value = None
        python_dataset.servicex.get_transform_status = AsyncMock(id="12345")
        python_dataset.servicex.get_transform_status.return_value = completed_status
        python_dataset.servicex.submit_transform = AsyncMock()
        python_dataset.download_files = AsyncMock()
        python_dataset.download_files.return_value = []

        signed_urls_only = False
        expandable_progress = ExpandableProgress()

        result = await python_dataset.submit_and_download(signed_urls_only, expandable_progress)
        assert result is not None
        assert result.request_id == "b8c508d0-ccf2-4deb-a1f7-65c839eebabf"
        cache.close()


@pytest.mark.asyncio
async def test_submit_and_download_cache_miss_overall_progress(python_dataset, completed_status):
    with tempfile.TemporaryDirectory() as temp_dir:
        python_dataset.current_status = None
        python_dataset.servicex = AsyncMock()
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        python_dataset.cache = cache
        python_dataset.configuration = config

        python_dataset.servicex = AsyncMock()
        python_dataset.cache.get_transform_by_hash = Mock()
        python_dataset.cache.get_transform_by_hash.return_value = None
        python_dataset.servicex.get_transform_status = AsyncMock(id="12345")
        python_dataset.servicex.get_transform_status.return_value = completed_status
        python_dataset.servicex.submit_transform = AsyncMock()
        python_dataset.download_files = AsyncMock()
        python_dataset.download_files.return_value = []

        signed_urls_only = False
        expandable_progress = ExpandableProgress(overall_progress=True)
        dataset_group = True

        result = await python_dataset.submit_and_download(signed_urls_only, expandable_progress,
                                                          dataset_group)
        assert result is not None
        assert result.request_id == "b8c508d0-ccf2-4deb-a1f7-65c839eebabf"
        cache.close()


@pytest.mark.asyncio
async def test_submit_and_download_no_result_format(python_dataset, completed_status):
    with tempfile.TemporaryDirectory() as temp_dir:
        python_dataset.current_status = None
        python_dataset.servicex = AsyncMock()
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        python_dataset.cache = cache
        python_dataset.configuration = config
        with pytest.raises(ValueError,
                           match=r"Unable to determine the result file format. "
                                 r"Use set_result_format method"):
            python_dataset.result_format = None
            python_dataset.servicex = AsyncMock()
            python_dataset.cache.get_transform_by_hash = Mock()
            python_dataset.cache.get_transform_by_hash.return_value = None
            python_dataset.servicex.get_transform_status = AsyncMock(id="12345")
            python_dataset.servicex.get_transform_status.return_value = completed_status
            python_dataset.servicex.submit_transform = AsyncMock()
            python_dataset.download_files = AsyncMock()
            python_dataset.download_files.return_value = []
            signed_urls_only = False
            expandable_progress = ExpandableProgress()
            await python_dataset.submit_and_download(signed_urls_only, expandable_progress)
        cache.close()


def test_set_title(python_dataset):
    python_dataset.set_title("Title Title")
    assert python_dataset.title == "Title Title"


@pytest.mark.asyncio
async def test_submit_and_download_cache_miss_signed_urls_only(python_dataset, completed_status):
    with tempfile.TemporaryDirectory() as temp_dir:
        python_dataset.current_status = None
        python_dataset.servicex = AsyncMock()
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        python_dataset.cache = cache
        python_dataset.configuration = config
        python_dataset.servicex = AsyncMock()
        python_dataset.cache.get_transform_by_hash = Mock()
        python_dataset.cache.get_transform_by_hash.return_value = None
        python_dataset.servicex.get_transform_status = AsyncMock(id="12345")
        python_dataset.servicex.get_transform_status.return_value = completed_status
        python_dataset.servicex.submit_transform = AsyncMock()
        python_dataset.download_files = AsyncMock()
        python_dataset.download_files.return_value = []

        signed_urls_only = True
        expandable_progress = ExpandableProgress()

        result = await python_dataset.submit_and_download(signed_urls_only, expandable_progress)
        assert result is not None
        assert result.request_id == "b8c508d0-ccf2-4deb-a1f7-65c839eebabf"
        cache.close()


@pytest.mark.asyncio
async def test_submit_and_download_cache_files_request_urls(python_dataset, transformed_result):
    with tempfile.TemporaryDirectory() as temp_dir:
        python_dataset.current_status = None
        python_dataset.servicex = AsyncMock()
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        python_dataset.cache = cache
        python_dataset.configuration = config
        python_dataset.servicex = AsyncMock()
        python_dataset.cache.get_transform_by_hash = Mock()
        python_dataset.cache.get_transform_by_hash.return_value = transformed_result
        status = Mock(files=10, files_completed=5, files_failed=1, status=Status.complete)
        python_dataset.current_status = status
        python_dataset.retrieve_current_transform_status = AsyncMock(return_value=status)

        signed_urls_only = True
        expandable_progress = ExpandableProgress()
        result = await python_dataset.submit_and_download(signed_urls_only, expandable_progress)
        assert result is not None
        assert result.request_id == transformed_result.request_id
        cache.close()


@pytest.mark.asyncio
async def test_submit_and_download_cache_urls_request_files(python_dataset, transformed_result):
    with tempfile.TemporaryDirectory() as temp_dir:
        python_dataset.current_status = None
        python_dataset.servicex = AsyncMock()
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        python_dataset.cache = cache
        python_dataset.configuration = config
        python_dataset.servicex = AsyncMock()
        python_dataset.cache.get_transform_by_hash = Mock()
        transformed_result.signed_url_list = ["a.b.c.com"]
        transformed_result.file_list = []
        python_dataset.cache.get_transform_by_hash.return_value = transformed_result
        status = Mock(files=10, files_completed=5, files_failed=1, status=Status.complete)
        python_dataset.current_status = status
        python_dataset.retrieve_current_transform_status = AsyncMock(return_value=status)

        signed_urls_only = False
        expandable_progress = ExpandableProgress()
        result = await python_dataset.submit_and_download(signed_urls_only, expandable_progress)
        assert result is not None
        assert result.request_id == transformed_result.request_id
        cache.close()


@pytest.mark.asyncio
async def test_network_loss(python_dataset, transformed_result):
    with tempfile.TemporaryDirectory() as temp_dir:
        python_dataset.current_status = None
        python_dataset.servicex = AsyncMock()
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        python_dataset.cache = cache
        python_dataset.configuration = config
        python_dataset.download_path = Path("www.a.b.com")

        python_dataset.servicex = AsyncMock()
        status = Mock(files=10, files_completed=5, files_failed=1, status=Status.fatal)
        python_dataset.current_status = status

        python_dataset.cache.get_transform_by_hash = Mock()
        transformed_result.files = status.files
        python_dataset.cache.get_transform_by_hash.return_value = transformed_result

        python_dataset.servicex.get_transform_status = AsyncMock(id="12345")
        python_dataset.servicex.get_transform_status.return_value = status
        python_dataset.servicex.submit_transform = AsyncMock()
        python_dataset.download_files = AsyncMock()
        python_dataset.download_files.return_value = []

        signed_urls_only = False
        expandable_progress = ExpandableProgress()

        result = await python_dataset.submit_and_download(signed_urls_only, expandable_progress)
        assert result is not None
        assert result.request_id == "123-45-6789"
        cache.close()



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import asyncio
from typing import List, Optional, Union

from rich.progress import Progress

from servicex.query import Query
from servicex.expandable_progress import ExpandableProgress
from servicex.func_adl.func_adl_dataset import FuncADLQuery
from servicex.models import TransformedResults, ResultFormat
from make_it_sync import make_sync


DatasetGroupMember = Union[Query, FuncADLQuery]


class DatasetGroup:
    def __init__(self, datasets: List[DatasetGroupMember]):
        r"""
        A group of datasets that are to be transformed together. This is a convenience
        class to allow you to submit multiple datasets to a ServiceX instance and
        then wait for all of them to complete.

        :param datasets: List of transform request as dataset instances
        """
        self.tasks = []
        self.datasets = datasets

    def set_result_format(self, result_format: ResultFormat):
        r"""
        Set the result format for all the datasets in the group.

        :param result_format: ResultFormat instance
        """
        for dataset in self.datasets:
            dataset.set_result_format(result_format)
        return self

    async def as_signed_urls_async(
        self,
        display_progress: bool = True,
        provided_progress: Optional[Progress] = None,
    ) -> List[TransformedResults]:
        with ExpandableProgress(display_progress, provided_progress,
                                overall_progress=True) as progress:
            self.tasks = [
                d.as_signed_urls_async(provided_progress=progress, dataset_group=True)
                for d in self.datasets
            ]
            return await asyncio.gather(*self.tasks)

    as_signed_urls = make_sync(as_signed_urls_async)

    async def as_files_async(self,
                             display_progress: bool = True,
                             provided_progress: Optional[Progress] = None
                             ) -> List[TransformedResults]:
        with ExpandableProgress(display_progress, provided_progress) as progress:
            self.tasks = [
                d.as_files_async(provided_progress=progress)
                for d in self.datasets
            ]
            return await asyncio.gather(*self.tasks)

    as_files = make_sync(as_files_async)



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from enum import Enum
from typing import Union, Optional, Callable, List
from pydantic import BaseModel, Field, root_validator, constr, validator

from servicex.dataset_identifier import RucioDatasetIdentifier, FileListDataset
from servicex.func_adl import func_adl_dataset


class Sample(BaseModel):
    Name: str
    Codegen: Optional[str]
    RucioDID: Optional[str]
    XRootDFiles: Optional[Union[str, List[str]]]
    NFiles: Optional[int] = Field(default=None)
    Function: Optional[Union[str, Callable]] = Field(default=None)
    Query: Optional[Union[str, func_adl_dataset.Query]] = Field(default=None)
    Tree: Optional[str] = Field(default=None)
    IgnoreLocalCache: bool = False

    class Config:
        arbitrary_types_allowed = True

    @property
    def dataset_identifier(self):
        if self.RucioDID:
            return RucioDatasetIdentifier(self.RucioDID, num_files=self.NFiles or 0)
        elif self.XRootDFiles:
            return FileListDataset(self.XRootDFiles)

    @root_validator
    def validate_did_xor_file(cls, values):
        """
        Ensure that only one of RootFile or RucioDID is specified.
        :param values:
        :return:
        """
        if values['XRootDFiles'] and values['RucioDID']:
            raise ValueError("Only specify one of XRootDFiles or RucioDID, not both.")
        if not values['XRootDFiles'] and not values['RucioDID']:
            raise ValueError("Must specify one of XRootDFiles or RucioDID.")
        return values

    @root_validator
    def validate_function_xor_query(cls, values):
        """
        Ensure that only one of Function or Query is specified.
        :param values:
        :return:
        """
        if values['Function'] and values['Query']:
            raise ValueError("Only specify one of Function or Query, not both.")
        if not values['Function'] and not values['Query']:
            raise ValueError("Must specify one of Function or Query.")
        return values


class General(BaseModel):
    class OutputFormatEnum(str, Enum):
        parquet = "parquet"
        root = "root-file"

    class DeliveryEnum(str, Enum):
        LocalCache = "LocalCache"
        SignedURLs = "SignedURLs"

    ServiceX: str = Field(..., alias="ServiceX")
    Codegen: str
    OutputFormat: Union[OutputFormatEnum,
        constr(regex='^(parquet|root-file)$')] = Field(default=OutputFormatEnum.root)  # NOQA F722

    Delivery: Union[DeliveryEnum, constr(regex='^(LocalCache|SignedURLs)$')] = Field(default=DeliveryEnum.LocalCache) # NOQA F722


class Definition(BaseModel):
    class Config:
        extra = "allow"  # Allow additional fields not defined in the model

    @root_validator
    def check_def_name(cls, values):
        """
        Ensure that the definition name is DEF_XXX format
        :param values:
        :return:
        """
        for field in values:
            if not field.startswith("DEF_"):
                raise ValueError(f"Definition key {field} does not meet the convention of DEF_XXX format") # NOQA E501
        return values


class ServiceXSpec(BaseModel):
    General: General
    Sample: List[Sample]
    Definition: Optional[Definition]

    @validator('Sample')
    def check_tree_property(cls, v, values):
        if 'General' in values and values['General'].Codegen != 'uproot':
            for sample in v:
                if 'Tree' in sample:
                    raise ValueError('"Tree" property is not allowed when codegen is not "uproot"')
        return v

    @root_validator(pre=False, skip_on_failure=True)
    def replace_definition(cls, values):
        """
        Replace the definition name with the actual definition value looking
        through the Samples and the General sections
        :param values:
        :return:
        """
        def replace_value_from_def(value, defs):
            """
            Replace the value with the actual definition value
            :param value:
            :param defs:
            :return:
            """
            if isinstance(value, dict):
                for field in value.keys():
                    value[field] = replace_value_from_def(value[field], defs)
            elif isinstance(value, str):
                if value.startswith("DEF_"):
                    if value in defs:
                        value = defs[value]
                    else:
                        raise ValueError(f"Definition {value} not found")
            return value

        if 'Definition' in values and values['Definition'] is not None:
            defs = values['Definition'].dict()
        else:
            defs = {}

        for sample_field in values['Sample']:
            replace_value_from_def(sample_field.__dict__, defs)

        replace_value_from_def(values['General'].__dict__, defs)

        return values



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import tempfile
from pathlib import Path, PurePath
from typing import List, Optional, Dict

from pydantic import BaseModel, Field, validator

import yaml


class Endpoint(BaseModel):
    endpoint: str
    name: str
    token: Optional[str]


class Configuration(BaseModel):
    api_endpoints: List[Endpoint]
    default_endpoint: Optional[str] = Field(alias="default-endpoint", default=None)
    cache_path: Optional[str] = Field(alias="cache-path", default=None)
    shortened_downloaded_filename: Optional[bool] = False

    @validator("cache_path", always=True)
    def expand_cache_path(cls, v):
        """
        Expand the cache path to a full path, and create it if it doesn't exist.
        Expand ${USER} to be the user name on the system. Works for windows, too.
        :param v:
        :return:
        """
        # create a folder inside the tmp directory if not specified in cache_path
        if not v:
            v = "/tmp/servicex_${USER}"

        s_path = os.path.expanduser(v)

        # If they have tried to use the USER or UserName as an expansion, and it has failed, then
        # translate it to maintain harmony across platforms.
        if "${USER}" in s_path and "UserName" in os.environ:
            s_path = s_path.replace("${USER}", os.environ["UserName"])
        if "${USER}" in s_path and "USER" in os.environ:
            s_path = s_path.replace("${USER}", os.environ["USER"])

        p_p = PurePath(s_path)
        if len(p_p.parts) > 1 and p_p.parts[1] == "tmp":
            p = Path(tempfile.gettempdir()) / Path(*p_p.parts[2:])
        else:
            p = Path(p_p)
        p.mkdir(exist_ok=True, parents=True)

        return p.as_posix()

    class Config:
        allow_population_by_field_name = True

    def endpoint_dict(self) -> Dict[str, Endpoint]:
        return {endpoint.name: endpoint for endpoint in self.api_endpoints}

    @classmethod
    def read(cls, config_path: Optional[str] = None):
        r"""
        Read configuration from .servicex or servicex.yaml file.
        :param config_path: If provided, use this as the path to the .servicex file.
                            Otherwise, search, starting from the current working directory
                            and look in enclosing directories
        :return: Populated configuration object
        """
        if config_path:
            yaml_config = cls._add_from_path(Path(config_path), walk_up_tree=False)
        else:
            yaml_config = cls._add_from_path(walk_up_tree=True)

        if yaml_config:
            return Configuration(**yaml_config)
        else:
            path_extra = f"in {config_path}" if config_path else ""
            raise NameError(
                "Can't find .servicex or servicex.yaml config file " + path_extra
            )

    @classmethod
    def _add_from_path(cls, path: Optional[Path] = None, walk_up_tree: bool = False):
        config = None
        if path:
            path.resolve()
            name = path.name
            dir = path.parent.resolve()
            alt_name = None
        else:
            name = ".servicex"
            alt_name = "servicex.yaml"
            dir = Path(os.getcwd())

        while True:
            f = dir / name  # user-defined path or .servicex
            if f.exists():
                with open(f) as config_file:
                    config = yaml.safe_load(config_file)
                    break

            if alt_name:
                f = dir / alt_name  # if neither option above, find servicex.yaml
                if f.exists():
                    with open(f) as config_file:
                        config = yaml.safe_load(config_file)
                        break

            if not walk_up_tree:
                break

            if dir == dir.parent:
                break

            dir = dir.parent

        return config



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from servicex import dataset_group
from servicex import models
from servicex import servicex_client
from servicex import dataset_identifier
from servicex.databinder_models import Sample, General, Definition, ServiceXSpec
from servicex.func_adl.func_adl_dataset import FuncADLQuery
from servicex.servicex_client import ServiceXClient, deliver
from .query import Query
from .models import ResultFormat, ResultDestination
from .dataset_group import DatasetGroup
from .dataset_identifier import RucioDatasetIdentifier, FileListDataset


__all__ = [
    "ServiceXClient",
    "Query",
    "DatasetGroup",
    "ResultFormat",
    "ResultDestination",
    "servicex_client",
    "dataset_group",
    "models",
    "dataset_identifier",
    "RucioDatasetIdentifier",
    "FileListDataset",
    "FuncADLQuery",
    "Sample",
    "General",
    "Definition",
    "ServiceXSpec",
    "deliver"
]



# Copyright (c) 2023, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import yaml
from pathlib import Path
from typing import Dict, Any
from rich import print
# from rich.panel import Panel


class OutputHandler:
    def __init__(self, updated_config: Dict[str, Any]):
        self._config = updated_config
        self.out_dict = self._initialize_out_dict()
        self._out_path = self._get_out_dir()

    def _initialize_out_dict(self):
        return {"general_info": {}, "samples": {}}

    def _get_out_dir(self):
        """
        Get output directory
        """
        if 'OutputDirectory' in self._config['General'].keys():
            self.output_path = Path(
                self._config['General']['OutputDirectory']
            ).absolute()
            self.output_path.mkdir(parents=True, exist_ok=True)
        else:
            self.output_path = Path().absolute()

    def update_out_dict(self, delivery: str, results):
        """
        Update output dict
        """
        if delivery == "objectstore":
            self.out_dict['samples'].update(
                {results.title: results.signed_url_list}
            )
        else:
            self.out_dict['samples'].update(
                {results.title: results.file_list}
            )

    def write_out_dict(self):
        file_out_paths = \
            (f"{self.output_path}/"
             f"{self._config['General']['OutFilesetName']}.yml")
        with open(file_out_paths, 'w') as f:
            yaml.dump(self.out_dict, f, default_flow_style=False)

    def copy_to_destination(self, delivery: str, results):
        sample_name = f"[bold red]{results.title}[/bold red]"
        if delivery == "objectstore":
            if len(results.signed_url_list):
                print(f"{sample_name} is available at the object store")
            else:
                print(f"Failed to deliver {sample_name}")
        else:
            if len(results.file_list):
                print(f"{sample_name} is delivered to the local cache directory")
            else:
                print(f"Failed to deliver {sample_name}")



# Copyright (c) 2023, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Any, Dict
# from rich.progress import Progress, TextColumn, BarColumn, MofNCompleteColumn, \
#     TimeRemainingColumn
from rich import print
import asyncio
import nest_asyncio

from servicex.databinder.databinder_requests import DataBinderRequests
from servicex.databinder.databinder_outputs import OutputHandler
from servicex.expandable_progress import ExpandableProgress

nest_asyncio.apply()


class DataBinderDeliver:
    """
    a
    """

    def __init__(self, updated_config: Dict[str, Any]) -> None:
        self._config = updated_config

        databinder_requests = DataBinderRequests(self._config)
        self._requests = databinder_requests.get_requests()
        self._endpoint = databinder_requests._client \
            .endpoints[updated_config['General']['ServiceX']].endpoint
        self._output_handler = OutputHandler(self._config)

    async def deliver_and_copy(self, req, progress):
        if req['delivery'] == "objectstore":
            results = await req['ds_query'].as_signed_urls_async(provided_progress=progress)
        else:
            results = await req['ds_query'].as_files_async(provided_progress=progress)

        # Update out_dict
        self._output_handler.update_out_dict(
            delivery=req['delivery'],
            results=results
        )

        #
        self._output_handler.copy_to_destination(
            delivery=req['delivery'],
            results=results
        )

    async def get_data(self):
        print(f"Deliver via ServiceX endpoint: [link=self._endpoint]{self._endpoint}[/link]")

        tasks = []
        with ExpandableProgress() as progress:
            for req in self._requests:
                tasks.append(self.deliver_and_copy(req, progress))

            for f in asyncio.as_completed(tasks):
                await f
                # print(value)

        self._output_handler.write_out_dict()

        return self._output_handler.out_dict
